import logging
import hikaru
from typing import Dict, Any, Optional
from pydantic import BaseModel

from .autogenerated.events import KIND_TO_EVENT_CLASS
from .autogenerated.models import get_api_version
from .model_not_found_exception import ModelNotFoundException
from ..helper import prefix_match, exact_match
from ...core.playbooks.base_trigger import BaseTrigger
from ...core.playbooks.base_trigger import TriggerEvent
from ...core.model.k8s_operation_type import K8sOperationType
from ...core.model.events import ExecutionBaseEvent
from ...core.reporting.base import Finding


class IncomingK8sEventPayload(BaseModel):
    """
    The format of incoming payloads containing kubernetes events. This is mostly used for deserialization.
    """

    operation: str
    kind: str
    apiVersion: str = ""
    clusterUid: str
    description: str
    obj: Dict[Any, Any]
    oldObj: Optional[Dict[Any, Any]]


class K8sTriggerEvent(TriggerEvent):
    k8s_payload: IncomingK8sEventPayload

    def get_event_name(self) -> str:
        return K8sTriggerEvent.__name__


class K8sBaseTrigger(BaseTrigger):
    kind: str
    operation: K8sOperationType = None
    name_prefix: str = None
    namespace_prefix: str = None

    def get_trigger_event(self):
        return K8sTriggerEvent.__name__

    def should_fire(self, event: TriggerEvent):
        if not isinstance(event, K8sTriggerEvent):
            return False

        k8s_trigger_event = K8sTriggerEvent(**event.dict())
        k8s_payload = k8s_trigger_event.k8s_payload
        if self.kind != "Any" and self.kind != k8s_payload.kind:
            return False

        if not exact_match(self.operation, K8sOperationType(k8s_payload.operation)):
            return False

        meta = k8s_payload.obj.get("metadata", {})
        if not prefix_match(self.name_prefix, meta.get("name", "")):
            return False

        if not prefix_match(self.namespace_prefix, meta.get("namespace", "")):
            return False

        return True

    @classmethod
    def __parse_kubernetes_objs(cls, k8s_payload: IncomingK8sEventPayload):
        model_class = get_api_version(k8s_payload.apiVersion).get(k8s_payload.kind)
        if model_class is None:
            msg = f"classes for kind {k8s_payload.kind} cannot be found. skipping. description {k8s_payload.description}"
            logging.error(msg)
            raise ModelNotFoundException(msg)

        obj = hikaru.from_dict(k8s_payload.obj, cls=model_class)
        old_obj = None
        if k8s_payload.oldObj is not None:
            old_obj = hikaru.from_dict(k8s_payload.oldObj, cls=model_class)
        return obj, old_obj

    def build_execution_event(
        self, event: K8sTriggerEvent, findings: Dict[str, Finding]
    ) -> Optional[ExecutionBaseEvent]:
        event_class = KIND_TO_EVENT_CLASS.get(event.k8s_payload.kind)
        if event_class is None:
            logging.error(
                f"classes for kind {event.k8s_payload.kind} cannot be found. skipping. description {event.k8s_payload.description}"
            )
            return None
        (obj, old_obj) = self.__parse_kubernetes_objs(event.k8s_payload)
        operation_type = K8sOperationType(event.k8s_payload.operation)
        return event_class(
            findings=findings,
            operation=operation_type,
            description=event.k8s_payload.description.replace("\n", ""),
            obj=obj,
            old_obj=old_obj,
        )
